/*******************************************************************************
   File:
      CLOCK.C

   Purpose:
      Chess Timer for the Pilot

   $Revision: 1.35 $
   $Header: F:\devel\pilot\clock/RCS/clock.c,v 1.35 1997/09/14 00:09:51 eshayne Exp $
   $Log: clock.c,v $
 * Revision 1.35  1997/09/14  00:09:51  eshayne
 * if hardware buttons are enabled, either button can start the initial
 * countdown
 *
 * Revision 1.34  1997/09/13  22:49:21  eshayne
 * hide software buttons if hardware buttons are enabled
 *
 * Revision 1.33  1997/09/13  22:32:45  eshayne
 * added confirmation if enable hardware buttons
 *
 * Revision 1.32  1997/09/13  22:24:05  eshayne
 * added preferences option to enable or disable hardware buttons
 *
 * Revision 1.31  1997/09/13  21:28:23  eshayne
 * added hardware buttons 1 & 4 to be substitutable for the
 * "switch" buttons
 *
 * Revision 1.30  1997/09/13  21:10:57  eshayne
 * made two "switch" buttons, one for each player
 *
 * Revision 1.29  1997/09/08  00:55:42  eshayne
 * merged common functionality from inital setup and Reset() into two
 * helper functions, SetInitialValues() and DrawInitialFields() for
 * before and after the form is drawn, respectively.
 *
 * Revision 1.28  1997/09/07  19:11:25  eshayne
 * added GNU Public License to source files
 *
 * Revision 1.27  1997/09/07  18:33:40  eshayne
 * replaced global preference vars with one instance of preference struct
 *
 * Revision 1.26  1997/09/07  18:09:34  eshayne
 * linked creator-id in code to creator-id in Makefile
 *
 * Revision 1.25  1997/09/07  18:02:26  eshayne
 * made preferences persistent
 *
 * Revision 1.24  1997/09/01  23:41:59  eshayne
 * added optional pause button
 *
 * Revision 1.23  1997/09/01  20:58:08  eshayne
 * made switching routine more accurate (I think) by capturing amount
 * of time left before next decrement before switching over - so next
 * time we can only wait that long before decrementing again. I'm not
 * convinced it works, but it's pretty much impossible to test...
 *
 * Revision 1.22  1997/09/01  19:34:02  eshayne
 * made button switch label between "Start" and "Switch," appropriately
 *
 * Revision 1.21  1997/09/01  19:09:39  eshayne
 * added popup/beep when time is out for one player
 *
 * Revision 1.20  1997/09/01  18:55:00  eshayne
 * added validation for preferences screen
 *
 * Revision 1.19  1997/09/01  18:20:44  eshayne
 * preset preferences screen with current values
 *
 * Revision 1.18  1997/09/01  17:24:36  eshayne
 * removed "Fischer Time?" checkbox
 * made preferences actually count if user hits "OK"
 *
 * Revision 1.17  1997/09/01  16:37:26  eshayne
 * added confirmation dialog on reset
 *
 * Revision 1.16  1997/09/01  16:31:20  eshayne
 * added Reset option to menu
 *
 * Revision 1.15  1997/09/01  16:08:30  eshayne
 * added Fischer time
 *
 * Revision 1.14  1997/09/01  14:43:52  eshayne
 * moved all TimeCounter-related stuff to counter.h/.c
 *
 * Revision 1.13  1997/09/01  14:31:12  eshayne
 * moved global variable definitions to clockGlobals.c
 *
 * Revision 1.12  1997/09/01  14:21:44  eshayne
 * form now initially shows initial counter values before first "switch"
 *
 * Revision 1.11  1997/08/31  21:50:51  eshayne
 * added Settings menu option
 * added Preferences form to real app
 * Preferences form settings still ignored, though
 *
 * Revision 1.10  1997/08/31  19:22:15  eshayne
 * added leading 0's
 *
 * Revision 1.9  1997/08/31  19:12:38  eshayne
 * added "About" dialog
 * added menu
 * added resources for preferences dialog, but no hooks yet
 *
 * Revision 1.8  1997/08/31  02:34:55  eshayne
 * added comments
 *
 * Revision 1.7  1997/08/31  02:25:51  eshayne
 * extern-ized globals
 *
 * Revision 1.6  1997/08/31  02:15:58  eshayne
 * added form handler
 *
 * Revision 1.5  1997/08/20  02:39:56  eshayne
 * added second player's timer and functionality to switch between the two
 * instead of just freezing
 *
 * Revision 1.4  1997/08/20  02:04:57  eshayne
 * adjusted to use clockResDefs for resource IDs
 *
 * Revision 1.3  1997/08/20  01:42:16  eshayne
 * moved TimeCounter to clockGlobals.h
 * moved stuff into helper functions for clarity
 *
 * Revision 1.2  1997/08/17  17:53:16  eshayne
 * added comment header block
 *

    Chess Timer
    Copyright (C) 1997  Ethan Shayne

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*******************************************************************************/

#include <Common.h>
#include <System/SysAll.h>
#include <UI/UIAll.h>

#include <counter.h>
#include <clockGlobals.h>
#include <clockResDefs.h>


/* standard application functions */
static int StartApplication();
static void EventLoop();
static void StopApplication();

/* form event handlers */
Boolean ClockMainFormHandler(EventPtr pEvent);
Boolean ClockPreferencesFormHandler(EventPtr pEvent);

/* helper functions */
static void Reset();
static Boolean GetPreferencesFromForm(FormPtr pForm);
static void PrefillPreferencesData(FormPtr pForm);
static void SetFieldText(FieldPtr pField, CharPtr pszText);
static Boolean ValidatePreferences(CharPtr pszInitMinutes,
				   CharPtr pszInitSeconds,
				   CharPtr pszFischer,
				   CharPtr pszAccuracy,
				   short sPossiblePauseEnabled,
				   short sPossibleHardwareButtonsEnabled,
				   unsigned long* pulInitMinutes,
				   unsigned long* pulInitSeconds,
				   unsigned long* pulFischer,
				   unsigned long* pulAccuracy,
				   int* piPauseEnabled,
				   int* piHardwareButtonsEnabled);
static void GetPreferences();
static void SavePreferences();
static void SetInitialValues();
static void DrawInitialFields(FormPtr pForm, const Boolean boolInitial);
static void ToggleSwitchButtons(const enum SwitchButtonToggles eToggleType);


/* PilotMain */
DWord PilotMain (Word cmd, Ptr cmdPBP, Word launchFlags)
{
  Word error;
    
  if (cmd == sysAppLaunchCmdNormalLaunch) {
    error = StartApplication();
    if (error) return error;

    EventLoop();

    StopApplication();
  }

  return 0;
}


static int StartApplication()
{
  FrmGotoForm(ES_CLOCK_FORMID);
}


/* Event Loop */
static void EventLoop()
{
  EventType event;
  short err;
  int formId;
  FormPtr form;

  do {
    /* get the next event in the queue */
    EvtGetEvent(&event, prefs.ulWait);

    /* let the system have first shot at it */
    /* ... but only if it's not an event indicating that
       hardware button 1 or 4 has been pressed, because we'll
       handle that (interpreted as a "switch" button press) */
    /* ... but only intercept it before the system gets to it
       if our preferences say that hardware buttons are enabled */
    if (!prefs.iHardwareButtonsEnabled ||
	!((event.eType == keyDownEvent)
	  && ((event.data.keyDown.chr == hard1Chr)
	      || (event.data.keyDown.chr == hard4Chr))))
      {
	if (SysHandleEvent(&event)) continue;
      }
    if (MenuHandleEvent(NULL, &event, &err)) continue;

    switch (event.eType)
      {
	/* if we're loading a new form, load it and set the event handler */
      case frmLoadEvent:
	formId = event.data.frmLoad.formID;
	form = FrmInitForm(formId);
	FrmSetActiveForm(form);
	
	switch (formId)
	  {
	    /* if it's the main form, use the main form handler */
	  case ES_CLOCK_FORMID:
	    FrmSetEventHandler(form, ClockMainFormHandler);
	    break;

	  case ES_CLOCK_PREF_FORMID:
	    FrmSetEventHandler(form, ClockPreferencesFormHandler);
	    break;
	  } 

	break;
      }


    /* if we've got everything set up (the form's been opened and drawn)
       then every loop we should whichever timer is currently active */
    /* ... but only if we're on the main form */
    /* ... and only if we've actually started (i.e. someone's pressed
       the button) */
    if (iInitialized &&
	(FrmGetActiveFormID() == ES_CLOCK_FORMID) &&
	(iFirstPlayerStarted || iSecondPlayerStarted) &&
	!iPaused)
      {
	if (iCurrentPlayer == 1) {
	  if (!UpdateTimer(&ulNextCountTick,
			   &counter1,
			   pMinutesField1,
			   pSecondsField1))
	    {
	      FrmCustomAlert(ES_CLOCK_GAMEOVER_ALERT, "1", "", "");
	      Reset();
	    }
	}
	else if (iCurrentPlayer == 2) {
	  if (!UpdateTimer(&ulNextCountTick,
			   &counter2,
			   pMinutesField2,
			   pSecondsField2))
	    {
	      FrmCustomAlert(ES_CLOCK_GAMEOVER_ALERT, "2", "", "");
	      Reset();
	    }
	}
      }

    /* if nothing else handled the event, dispatch it to the current event
       handler */
    FrmDispatchEvent(&event);

    /*keep going until the event is a "stop" event (application terminated) */
  } while (event.eType != appStopEvent);
}


static void StopApplication()
{
}


/* event handler for the main form */
Boolean ClockMainFormHandler(EventPtr pEvent)
{
  Boolean boolHandled = false;
  FormPtr form = FrmGetActiveForm();
  Word wObjIndex;
  ControlPtr pButton;
  ControlPtr pPauseButton;
  unsigned long ulOldNextTick = 0;

  switch (pEvent->eType)
    {
    case frmOpenEvent:
      /* initalize the initial values of the counters, labels, etc. */
      SetInitialValues();

      /* draw the form */
      FrmDrawForm(form);

      /* set up some global field pointers for later use, just so
	 we don't need to keep finding them every time */
      wObjIndex = FrmGetObjectIndex(form, ES_CLOCK_PLAYER1_MINUTES_FIELD);
      pMinutesField1 = FrmGetObjectPtr(form, wObjIndex);
      wObjIndex = FrmGetObjectIndex(form, ES_CLOCK_PLAYER1_SECONDS_FIELD);
      pSecondsField1 = FrmGetObjectPtr(form, wObjIndex);
      wObjIndex = FrmGetObjectIndex(form, ES_CLOCK_PLAYER2_MINUTES_FIELD);
      pMinutesField2 = FrmGetObjectPtr(form, wObjIndex);
      wObjIndex = FrmGetObjectIndex(form, ES_CLOCK_PLAYER2_SECONDS_FIELD);
      pSecondsField2 = FrmGetObjectPtr(form, wObjIndex);

      /* draw in some of the initial actual field data */
      DrawInitialFields(form, true);

      iInitialized = 1;
      boolHandled = true;
      break;
      
    /* if the event was a press of a button on the form... */
    case ctlSelectEvent:
      /* if the button was our "switch" button... */
      /* there are a few distinctions to make at this point - this
	 section of code should be triggered if: it's player 1's
	 turn and button 1 was pressed; it's player 2's turn and
	 button 2 was pressed; or the start button was pressed. */
      if (((pEvent->data.ctlEnter.controlID == ES_CLOCK_SWITCH_BUTTON1) &&
	   (iCurrentPlayer == 1)) ||
	  ((pEvent->data.ctlEnter.controlID == ES_CLOCK_SWITCH_BUTTON2) &&
	   (iCurrentPlayer == 2)) ||
	  (pEvent->data.ctlEnter.controlID == ES_CLOCK_START_BUTTON))
	{
	  /* if we're paused, ignore the switch request */
	  if (!iPaused) {
	    ulOldNextTick = ulNextCountTick;
	  
	    /* if the player we're about to switch to has been switched
	       away from yet, set the next decrement marker based on how
	       much time was left last time we switched away from this player.
	       otherwise just use sysTicksPerSecond. */
	    if (((iCurrentPlayer == 1) && (iSecondPlayerStarted)) ||
		((iCurrentPlayer == 2) && (iFirstPlayerStarted)))
	      {
		ulNextCountTick = TimGetTicks() + ulLastLeftoverTicks;
	      }
	    else
	      ulNextCountTick = TimGetTicks() + sysTicksPerSecond;
	    
	    /* if we're switching (and not just starting) - remember how
	       much time was leftover before the next counter decrement
	       so we can take that into account when we switch back */
	    if (ulNextCountTick)
	    ulLastLeftoverTicks = ulOldNextTick - TimGetTicks();
	    
	    /* change the current player/counter */
	    if (iCurrentPlayer == 1) {
	      iCurrentPlayer = 2;
	      /* add fischer time if this isn't this player's first move */
	      if (prefs.ulFischer && iSecondPlayerStarted) {
		IncrementCounter(&counter2, prefs.ulFischer);
		DrawTimer(&counter2,
			  pMinutesField2,
			  pSecondsField2);
	      }
	      /* if this was this player's first move, next one won't be */
	      if (!iSecondPlayerStarted)
		iSecondPlayerStarted = 1;
	    }
	    else {
	      iCurrentPlayer = 1;
	      /* add fischer time if this isn't this player's first move */
	      if (prefs.ulFischer && iFirstPlayerStarted) {
		IncrementCounter(&counter1, prefs.ulFischer);
		DrawTimer(&counter1,
			  pMinutesField1,
			  pSecondsField1);
	      }
	      /* if this was this player's first move, next one won't be */
	      if (!iFirstPlayerStarted)
		iFirstPlayerStarted = 1;
	    }
	    
	    
	    /* if this was the first button press, switch the button's
	       label to "Switch" */
	    /* ... but only if hardware buttons are not enabled */
	    if (!prefs.iHardwareButtonsEnabled &&
		(!iFirstPlayerStarted || !iSecondPlayerStarted))
	      {
		ToggleSwitchButtons(SWITCH);
	      }
	  }
	}
      
      else if (pEvent->data.ctlEnter.controlID == ES_CLOCK_PAUSE_BUTTON) {
	/* pause the countdown if the pause button was pressed */
	/* if we haven't started yet, ignore the pause request */
	if (iFirstPlayerStarted || iSecondPlayerStarted) {

	  /* see if we're pausing or un-pausing */
	  if (!iPaused) {
	    /* set the flag so the next event loop won't continue counting */
	    iPaused = 1;
	    
	    /* remember how far we were from a counter decrement... */
	    ulLastLeftoverPauseTicks = ulNextCountTick - TimGetTicks();
	  }
	  else {
	    /* if we're un-pausing... */
	    iPaused = 0;
	    ulNextCountTick = TimGetTicks() + ulLastLeftoverPauseTicks;
	  }

	}

      }

      boolHandled = true;
      break;

    case menuEvent:
      switch (pEvent->data.menu.itemID)
	{
	case ES_CLOCK_MENU_ABOUT:
	  FrmAlert(ES_CLOCK_ABOUT_ALERT);
	  boolHandled = true;
	  break;

	case ES_CLOCK_MENU_SETTINGS:
	  FrmPopupForm(ES_CLOCK_PREF_FORMID);
	  boolHandled = true;
	  break;

	case ES_CLOCK_MENU_RESET:
	  if (!FrmAlert(ES_CLOCK_CONFIRM_RESET_ALERT))
	    Reset();
	  boolHandled = true;
	  break;
	}

    /* if one of the hardware buttons was pressed... */
    case keyDownEvent:
      /* in the somewhat "special" situation where we haven't started
	 yet, but we are using the hardware buttons, *either* hardware
	 button should be interpretable as "start" - in other words,
	 just as if when hardware buttons are disabled, the user had
	 hit the "start" button. */
      if ((!iFirstPlayerStarted && !iSecondPlayerStarted) &&
	  ((pEvent->data.keyDown.chr == hard1Chr) ||
	   (pEvent->data.keyDown.chr == hard4Chr)))
	{
	  wObjIndex = FrmGetObjectIndex(form, ES_CLOCK_START_BUTTON);
	  pButton = FrmGetObjectPtr(form, wObjIndex);
	  CtlHitControl(pButton);

	  boolHandled = true;
	}

      /* otherwise, the hardware buttons should be interpreted as
	 either player 1's button or player 2's button */
      else {
	/* if it was the first hardware button (the calendar one),
	   pretend the user hit switch button #1 */
	if (pEvent->data.keyDown.chr == hard1Chr) {
	  wObjIndex = FrmGetObjectIndex(form, ES_CLOCK_SWITCH_BUTTON1);
	  pButton = FrmGetObjectPtr(form, wObjIndex);
	  CtlHitControl(pButton);
	  
	  boolHandled = true;
	}
	
	/* if it was the fourth hardware button (the memopad one),
	   pretend the user hit switch button #2 */
	if (pEvent->data.keyDown.chr == hard4Chr) {
	  wObjIndex = FrmGetObjectIndex(form, ES_CLOCK_SWITCH_BUTTON2);
	  pButton = FrmGetObjectPtr(form, wObjIndex);
	  CtlHitControl(pButton);
	  
	  boolHandled = true;
	}
      }
    }

  return boolHandled;
}


Boolean ClockPreferencesFormHandler(EventPtr pEvent)
{
  Boolean boolHandled = false;
  FormPtr form = FrmGetActiveForm();

  switch (pEvent->eType)
    {
    case frmOpenEvent:
      /* draw the form */
      PrefillPreferencesData(form);
      FrmDrawForm(form);
      boolHandled = true;
      break;

    case ctlSelectEvent:
      if (pEvent->data.ctlEnter.controlID == ES_CLOCK_PREFS_OK_BUTTON) {
	/* take the values from the form and set the global variables */
	if (GetPreferencesFromForm(form)) {
	  FrmReturnToForm(ES_CLOCK_FORMID);
	  Reset();
	}
	boolHandled = true;
      }
      else if (pEvent->data.ctlEnter.controlID == ES_CLOCK_PREFS_CANCEL_BUTTON) {
	FrmReturnToForm(ES_CLOCK_FORMID);
	boolHandled = true;
      }

      break;
    }

  return boolHandled;
}


static void Reset()
{
  FormPtr pForm = FrmGetActiveForm();

  SetInitialValues();
  DrawInitialFields(pForm, false);
}


static Boolean GetPreferencesFromForm(FormPtr pForm)
{
  Word wObjIndex;
  FieldPtr pInitMinutes;
  FieldPtr pInitSeconds;
  FieldPtr pFischerSeconds;
  FieldPtr pAccuracy;
  ControlPtr pPauseEnabled;
  ControlPtr pHardwareEnabled;

  CharPtr pszInitMinutes = NULL;
  CharPtr pszInitSeconds = NULL;
  CharPtr pszFischerSeconds = NULL;
  CharPtr pszAccuracy = NULL;
  short sTempPauseEnabled = 0;
  short sTempHardwareEnabled = 0;

  Boolean boolReturn;

  wObjIndex = FrmGetObjectIndex(pForm, ES_CLOCK_PREFS_INIT_MIN_FIELD);
  pInitMinutes = FrmGetObjectPtr(pForm, wObjIndex);
  wObjIndex = FrmGetObjectIndex(pForm, ES_CLOCK_PREFS_INIT_SEC_FIELD);
  pInitSeconds = FrmGetObjectPtr(pForm, wObjIndex);
  wObjIndex = FrmGetObjectIndex(pForm, ES_CLOCK_PREFS_FISCHER_SECONDS_FIELD);
  pFischerSeconds = FrmGetObjectPtr(pForm, wObjIndex);
  wObjIndex = FrmGetObjectIndex(pForm, ES_CLOCK_PREFS_ACCURACY_FIELD);
  pAccuracy = FrmGetObjectPtr(pForm, wObjIndex);
  wObjIndex = FrmGetObjectIndex(pForm, ES_CLOCK_PREFS_PAUSE_ENABLED_BOX);
  pPauseEnabled = FrmGetObjectPtr(pForm, wObjIndex);
  wObjIndex = FrmGetObjectIndex(pForm, ES_CLOCK_PREFS_HARDWARE_ENABLED_BOX);
  pHardwareEnabled = FrmGetObjectPtr(pForm, wObjIndex);
  
  pszInitMinutes = FldGetTextPtr(pInitMinutes);
  pszInitSeconds = FldGetTextPtr(pInitSeconds);
  pszFischerSeconds = FldGetTextPtr(pFischerSeconds);
  pszAccuracy = FldGetTextPtr(pAccuracy);
  sTempPauseEnabled = CtlGetValue(pPauseEnabled);
  sTempHardwareEnabled = CtlGetValue(pHardwareEnabled);
  
  boolReturn = ValidatePreferences(pszInitMinutes,
				   pszInitSeconds,
				   pszFischerSeconds,
				   pszAccuracy,
				   sTempPauseEnabled,
				   sTempHardwareEnabled,
				   &prefs.ulStartMinutes,
				   &prefs.ulStartSeconds,
				   &prefs.ulFischer,
				   &prefs.ulWait,
				   &prefs.iPauseEnabled,
				   &prefs.iHardwareButtonsEnabled);

  if (boolReturn)
    SavePreferences();

  return boolReturn;
}


static void PrefillPreferencesData(FormPtr pForm)
{
  static char pszInitMinutes[5];
  static char pszInitSeconds[5];
  static char pszFischer[5];
  static char pszAccuracy[5];

  Word wObjIndex;
  FieldPtr pInitMinutes;
  FieldPtr pInitSeconds;
  FieldPtr pFischerSeconds;
  FieldPtr pAccuracy;
  ControlPtr pPauseEnabled;
  ControlPtr pHardwareEnabled;

  wObjIndex = FrmGetObjectIndex(pForm, ES_CLOCK_PREFS_INIT_MIN_FIELD);
  pInitMinutes = FrmGetObjectPtr(pForm, wObjIndex);
  wObjIndex = FrmGetObjectIndex(pForm, ES_CLOCK_PREFS_INIT_SEC_FIELD);
  pInitSeconds = FrmGetObjectPtr(pForm, wObjIndex);
  wObjIndex = FrmGetObjectIndex(pForm, ES_CLOCK_PREFS_FISCHER_SECONDS_FIELD);
  pFischerSeconds = FrmGetObjectPtr(pForm, wObjIndex);
  wObjIndex = FrmGetObjectIndex(pForm, ES_CLOCK_PREFS_ACCURACY_FIELD);
  pAccuracy = FrmGetObjectPtr(pForm, wObjIndex);
  wObjIndex = FrmGetObjectIndex(pForm, ES_CLOCK_PREFS_PAUSE_ENABLED_BOX);
  pPauseEnabled = FrmGetObjectPtr(pForm, wObjIndex);
  wObjIndex = FrmGetObjectIndex(pForm, ES_CLOCK_PREFS_HARDWARE_ENABLED_BOX);
  pHardwareEnabled = FrmGetObjectPtr(pForm, wObjIndex);
  
  StrIToA(pszInitMinutes, prefs.ulStartMinutes);
  StrIToA(pszInitSeconds, prefs.ulStartSeconds);
  StrIToA(pszFischer, prefs.ulFischer);
  StrIToA(pszAccuracy, prefs.ulWait);

  SetFieldText(pInitMinutes, pszInitMinutes);
  SetFieldText(pInitSeconds, pszInitSeconds);
  SetFieldText(pFischerSeconds, pszFischer);
  SetFieldText(pAccuracy, pszAccuracy);
  CtlSetValue(pPauseEnabled, prefs.iPauseEnabled);
  CtlSetValue(pHardwareEnabled, prefs.iHardwareButtonsEnabled);
}


/* Function puts the supplied text into the supplied field. The supplied
   field can be an editable field (so FieldSetTextPtr is not permissible).
   Thanks to Ken Hancock for describing how to write this function. */
static void SetFieldText(FieldPtr pField, CharPtr pszNewText)
{
  VoidHand hFieldText;
  CharPtr pszActualText;
  
  /* first, free any existing data */
  FldFreeMemory(pField);
  
  /* allocate a new handle for the new text */
  hFieldText = MemHandleNew(StrLen(pszNewText) + 1);

  /* copy the supplied data into the new handle */
  /* get a pointer to the actual data held by the handle (& lock it) */
  pszActualText = MemHandleLock(hFieldText);
  StrCopy(pszActualText, pszNewText);
  
  /* set the field to use the new handle */
  FldSetTextHandle(pField, (Handle)hFieldText);

  /* unlock the handle */
  MemHandleUnlock(hFieldText);
}


static Boolean ValidatePreferences(CharPtr pszInitMinutes,
				   CharPtr pszInitSeconds,
				   CharPtr pszFischer,
				   CharPtr pszAccuracy,
				   short sPossiblePauseEnabled,
				   short sPossibleHardwareEnabled,
				   unsigned long* pulInitMinutes,
				   unsigned long* pulInitSeconds,
				   unsigned long* pulFischer,
				   unsigned long* pulAccuracy,
				   int* piPauseEnabled,
				   int* piHardwareEnabled)
{
  int iTempInitMinutes;
  int iTempInitSeconds;
  int iTempFischer;
  int iTempAccuracy;

  if (!pszInitMinutes || !pszInitSeconds || !pszFischer || !pszAccuracy) {
    if (!pszInitMinutes)
      FrmCustomAlert(ES_CLOCK_PREFS_INVALID_NULL_ALERT, "Initial Minutes", "" , "");
    if (!pszInitSeconds)
      FrmCustomAlert(ES_CLOCK_PREFS_INVALID_NULL_ALERT, "Initial Seconds", "" , "");
    if (!pszFischer)
      FrmCustomAlert(ES_CLOCK_PREFS_INVALID_NULL_ALERT, "Fischer Time", "" , "");
    if (!pszAccuracy)
      FrmCustomAlert(ES_CLOCK_PREFS_INVALID_NULL_ALERT, "Accuracy", "" , "");
    return false;
  }

  iTempInitMinutes = StrAToI(pszInitMinutes);
  iTempInitSeconds = StrAToI(pszInitSeconds);
  iTempFischer = StrAToI(pszFischer);
  iTempAccuracy = StrAToI(pszAccuracy);

  if ((iTempInitMinutes < 0) || (iTempInitMinutes > 59)) {
    FrmCustomAlert(ES_CLOCK_PREFS_INVALID_VALUE_ALERT, "Initial Minutes", "0" , "59");
    return false;
  }

  if ((iTempInitSeconds < 0) || (iTempInitSeconds > 59)) {
    FrmCustomAlert(ES_CLOCK_PREFS_INVALID_VALUE_ALERT, "Initial Seconds", "0" , "59");
    return false;
  }

  if (iTempFischer < 0) {
    FrmCustomAlert(ES_CLOCK_PREFS_INVALID_VALUE_ALERT, "Fischer Time", "0" , "");
    return false;
  }

  if ((iTempAccuracy < 1) || (iTempAccuracy > 100)) {
    FrmCustomAlert(ES_CLOCK_PREFS_INVALID_VALUE_ALERT, "Accuracy", "1" , "100");
    return false;
  }

  /* if (by checking against the current preferences) we see the user
     just chose to enable the hardware buttons, make sure they really
     want to do that */
  if (!prefs.iHardwareButtonsEnabled && sPossibleHardwareEnabled) {
    if (FrmAlert(ES_CLOCK_PREFS_HARDWARE_WARNING_ALERT))
      return false;
  }

  *pulInitMinutes = iTempInitMinutes;
  *pulInitSeconds = iTempInitSeconds;
  *pulFischer = iTempFischer;
  *pulAccuracy = iTempAccuracy;
  *piPauseEnabled = sPossiblePauseEnabled;
  *piHardwareEnabled = sPossibleHardwareEnabled;

  return true;
}


static void ToggleSwitchButtons(const enum SwitchButtonToggles eToggleType)
{
  Word wObjIndex;
  ControlPtr pButton;
  FormPtr pForm = FrmGetActiveForm();

  switch (eToggleType)
    {
    /* if eToggleType is START, then we want to erase the dual
       "switch" buttons and draw the single "start" button. */
    case START:
      /* erase the "switch" buttons */
      wObjIndex = FrmGetObjectIndex(pForm, ES_CLOCK_SWITCH_BUTTON1);
      pButton = FrmGetObjectPtr(pForm, wObjIndex);
      CtlHideControl(pButton);
      
      wObjIndex = FrmGetObjectIndex(pForm, ES_CLOCK_SWITCH_BUTTON2);
      pButton = FrmGetObjectPtr(pForm, wObjIndex);
      CtlHideControl(pButton);
      
      /* draw the "start" button */
      wObjIndex = FrmGetObjectIndex(pForm, ES_CLOCK_START_BUTTON);
      pButton = FrmGetObjectPtr(pForm, wObjIndex);
      CtlShowControl(pButton);
      
      break;

    /* if eToggleType is SWITCH, erase the "start" button and
       draw the "switch" buttons */
    case SWITCH:
      /* erase the "start" button */
      wObjIndex = FrmGetObjectIndex(pForm, ES_CLOCK_START_BUTTON);
      pButton = FrmGetObjectPtr(pForm, wObjIndex);
      CtlHideControl(pButton);
      
      /* draw the "switch" buttons */
      wObjIndex = FrmGetObjectIndex(pForm, ES_CLOCK_SWITCH_BUTTON1);
      pButton = FrmGetObjectPtr(pForm, wObjIndex);
      CtlShowControl(pButton);
      
      wObjIndex = FrmGetObjectIndex(pForm, ES_CLOCK_SWITCH_BUTTON2);
      pButton = FrmGetObjectPtr(pForm, wObjIndex);
      CtlShowControl(pButton);

      break;

    /* if eToggleType is HIDDEN, hide all the buttons */
    case HIDDEN:
      wObjIndex = FrmGetObjectIndex(pForm, ES_CLOCK_SWITCH_BUTTON1);
      pButton = FrmGetObjectPtr(pForm, wObjIndex);
      CtlHideControl(pButton);
      
      wObjIndex = FrmGetObjectIndex(pForm, ES_CLOCK_SWITCH_BUTTON2);
      pButton = FrmGetObjectPtr(pForm, wObjIndex);
      CtlHideControl(pButton);
      
      /* draw the "start" button */
      wObjIndex = FrmGetObjectIndex(pForm, ES_CLOCK_START_BUTTON);
      pButton = FrmGetObjectPtr(pForm, wObjIndex);
      CtlHideControl(pButton);
      
      break;      
    }
}


static void GetPreferences()
{
  /* if function returns false, prefs will be unaffected and defaults will
     remain in effect */
  /* note - CREATORID is a macro defined in the Makefile to match the creator
     id with which this app was compiled */
  PrefGetAppPreferences(CREATORID, 1, &prefs, sizeof(struct ClockPreferences));
}


static void SavePreferences()
{
  /* note - CREATORID is a macro defined in the Makefile to match the creator
     id with which this app was compiled */
  PrefSetAppPreferences(CREATORID, 1, &prefs, sizeof(struct ClockPreferences));
}


static void SetInitialValues()
{
  /* set the preferences based on the values stored in the
     Preferences database */
  GetPreferences();
  
  /* reset the timers */
  counter1.miMinutes = prefs.ulStartMinutes;
  counter1.miSeconds = prefs.ulStartSeconds;
  counter2.miMinutes = prefs.ulStartMinutes;
  counter2.miSeconds = prefs.ulStartSeconds;

  /* reset the flags indicating that we haven't started yet */  
  iFirstPlayerStarted = 0;
  iSecondPlayerStarted = 0;
  iCurrentPlayer = 2;
  ulNextCountTick = 0;
}


static void DrawInitialFields(FormPtr pForm, const Boolean boolInitial)
{
  Word wObjIndex;
  ControlPtr pPauseButton;
  ControlPtr pButton;

  /* enable or disable the pause button as per preferences */
  wObjIndex = FrmGetObjectIndex(pForm, ES_CLOCK_PAUSE_BUTTON);
  pPauseButton = FrmGetObjectPtr(pForm, wObjIndex);
  if (prefs.iPauseEnabled)
    CtlShowControl(pPauseButton);
  else
    /* if this is the first time the form is being drawn, there's
       no need to hide the pause button because it starts out nonusable */
    if (!boolInitial)
      CtlHideControl(pPauseButton);
  
  /* draw in the initial counter values */
  DrawTimer(&counter1,
	    pMinutesField1,
	    pSecondsField1);
  DrawTimer(&counter2,
	    pMinutesField2,
	    pSecondsField2);

  /* draw or hide the appropriate "switch" buttons */
  /* if the preferences indicate that the hardware buttons are
     enabled, hide all the "switch" buttons */
  if (prefs.iHardwareButtonsEnabled)
    ToggleSwitchButtons(HIDDEN);
  /* otherwise, set the "switch" buttons to show just the "start"
     button */
  else
    ToggleSwitchButtons(START);
}
